
bCastGCDBar = {}
bCastGCDBar.__index = bCastGCDBar

function bCastGCDBar.New()
	local gcd = {}
	setmetatable(gcd, bCastGCDBar)
	
	local config = bCast_profiles
	local curProf = bCast_currentProfile
	
	local context = UI.CreateContext("GCD Context")
	
	gcd.bar = UI.CreateFrame("Frame", "bCastGCDBar", context)
	gcd.bar.spark = UI.CreateFrame("Texture", "bCastGCDSpark", gcd.bar)
	
	gcd.width = config[curProf].player.width
	gcd.height = 4
	
	gcd.bar:SetWidth(gcd.width)
	gcd.bar:SetHeight(gcd.height)
	gcd.bar:SetBackgroundColor(0,0,0)
	gcd.bar:SetPoint("BOTTOMLEFT", UIParent, "TOPLEFT", config[curProf].player.xPos, config[curProf].player.yPos - 3) -- Set the GCD bar above player cast bar by default
	
	gcd.bar.spark:SetWidth(11)
	gcd.bar.spark:SetHeight(6)
	gcd.bar.spark:SetTexture("bCast", "media\\gcdSpark.png")
	gcd.bar.spark:SetPoint("CENTER", gcd.bar, "CENTERLEFT", 0, 0)
	
	gcd.bar:SetLayer(0)
	gcd.bar.spark:SetLayer(1)
	
	gcd.bar:SetVisible(false)
	
	return gcd
end

function bCastGCDBar:UpdatePosition()
	self.width = bCast_profiles[bCast_currentProfile].player.width
	self.bar:ClearAll()
	self.bar:SetWidth(self.width)
	self.bar:SetHeight(self.height)
	if bCast_profiles[bCast_currentProfile].player.showGCD == "above" then
		self.bar:SetPoint("BOTTOMLEFT", UIParent, "TOPLEFT", bCast_profiles[bCast_currentProfile].player.xPos, bCast_profiles[bCast_currentProfile].player.yPos - 3)
	else
		self.bar:SetPoint("TOPLEFT", UIParent, "TOPLEFT", bCast_profiles[bCast_currentProfile].player.xPos, bCast_profiles[bCast_currentProfile].player.yPos + bCast_profiles[bCast_currentProfile].player.height + 3)
	end
end

-- Cooldown is starting, find any ability with GCD on, set the bar visible and mark the starting time of the cd
function bCastGCDBar:CDStart(arg)
	for k,v in pairs(arg) do
		if v > 0.9 and v < 1.51 then -- assumes every CD between 0.9 and 1.51s is a GCD, rogues about 1s, about 1.5s for the rest
			-- GCD is something along the lines of 1.5000001192093 apparently, need to compare to a bigger value than 1.5
			self.tot = v -- total CD time
			self.sTime = Inspect.Time.Real() -- cast start time
			self.cdOn = true
			self.cAb = k -- track the GCD with this ability
			bCastUpdateStart(self)
			self:SetVisible(true)
			break
		end
	end
end


-- hiding the GCD bar based on time elapsed seems to be safer than relying on unreliable events
-- but this event is still needed to hide the GCD bar when player interrupts their cast during GCD, and the GCD along with it
--
-- At least in 1.6 this event won't always work, the table given sometimes includes only one ability, and if it's not the same we grabbed at start
-- there's not much that can be done with it.
function bCastGCDBar:CDEnd(arg)
	for k,v in pairs(arg) do
		if k == self.cAb then -- same ability as we started with, has its CD ending, which is < 1.51s
			self.cAb = nil
			self.cdOn = false	
			self:SetVisible(false)
			bCastUpdateStop(self)
			break
		end
	end
end

-- Moves the spark and if needed, hides the whole bar
function bCastGCDBar:Update()
	if self.cdOn then
		local pos = ( self.width * ( ( Inspect.Time.Real() - self.sTime ) / self.tot) )
		self.bar.spark:SetPoint("CENTER", self.bar, "CENTERLEFT", pos, 0)
		if pos > self.width then -- spark is outside the bar
			-- CD end doesn't always seem to get triggered, so use this method to reset instead
			self.cdOn = false
			self.cAb = nil
			self:SetVisible(false)
			bCastUpdateStop(self)
		end
	end
end

function bCastGCDBar:SetVisible(visible)
	self.bar:SetVisible(visible)
end