bCastVersion = "2.1c"

function bCastDefaults()

	local defaults = {
		player = {
			-- player cast bar settings
			xPos = 200,
			yPos = 200,
			width = 270,
			height = 32,
			colors = {
				bar_r = 0.2,
				bar_g = 0.4,
				bar_b = 1,
				text_r = 1,
				text_b = 1,
				text_g = 1
			},
			icon = "left", -- left/right/none
			texture = "smooth",
			font = "arial_round",
			fontSize = 13,
			maxLength = 27,
			pushBack = false,
			enabled = true,
			timeStyle = "remaining",  -- done/remaining, only available through slash commands
			showGCD = "above", -- none/above/below
		},
		["player.target"] = {
			-- target bar settings
			xPos = 200,
			yPos = 232,
			width = 240,
			height = 32,
			colors = {
				bar_r = 0.2,
				bar_g = 0.4,
				bar_b = 1,
				bar_ir = 0.2,
				bar_ig = 0.2,
				bar_ib = 0.2,
				text_r = 1,
				text_b = 1,
				text_g = 1
			},
			icon = "left", -- left/right/none
			texture = "smooth",
			font = "arial_round",
			fontSize = 13,
			maxLength = 23,
			enabled = true,
			timeStyle = "remaining",
		},
		["player.target.target"] = {
			-- target of target bar settings
			xPos = 200,
			yPos = 264,
			width = 240,
			height = 32,
			colors = {
				bar_r = 0.2,
				bar_g = 0.4,
				bar_b = 1,
				bar_ir = 0.2,
				bar_ig = 0.2,
				bar_ib = 0.2,
				text_r = 1,
				text_b = 1,
				text_g = 1
			},
			icon = "left", -- left/right/none
			texture = "smooth",
			font = "arial_round",
			fontSize = 13,
			maxLength = 23,
			enabled = true,
			timeStyle = "remaining",
		},
		focus = {
			-- focus bar settings
			xPos = 200,
			yPos = 296,
			width = 240,
			height = 32,
			colors = {
				bar_r = 0.2,
				bar_g = 0.4,
				bar_b = 1,
				bar_ir = 0.2,
				bar_ig = 0.2,
				bar_ib = 0.2,
				text_r = 1,
				text_b = 1,
				text_g = 1
			},
			icon = "left", -- left/right/none
			texture = "smooth",
			font = "arial_round",
			fontSize = 13,
			maxLength = 23,
			enabled = true,
			timeStyle = "remaining",
		},
	}

	return defaults
end
bCast_profiles = {}
bCast_profiles["default"] = bCastDefaults()
bCast_currentProfile = "default"

local cBars = {}
local enabledBars = {}
local locked = true
local gcd = nil

local updateQueue = {}

local CDStartFunc = nil
local CDEndFunc = nil
local gcdOn = false

local function CDStart(arg)
	gcd:CDStart(arg)
end

local function CDEnd(arg)
	gcd:CDEnd(arg)
end

local function GCDEnabler()
	if bCast_profiles[bCast_currentProfile].player.showGCD ~= "none" and not gcdOn then
		-- Enable gcd bar updates if needed
		table.insert(Event.Ability.Cooldown.Begin, CDStartFunc )
		table.insert(Event.Ability.Cooldown.End, CDEndFunc )
		gcdOn = true
	elseif gcdOn and bCast_profiles[bCast_currentProfile].player.showGCD == "none" then
		gcd:SetVisible(false)
		-- disable gcd bar updates
		local beginT = Event.Ability.Cooldown.Begin
		for i=1,#beginT do
			if CDStartFunc == beginT[i] then
				table.remove(Event.Ability.Cooldown.Begin, i)
				break
			end
		end
		local endT = Event.Ability.Cooldown.End
		for i=1,#endT do
			if CDEndFunc == endT[i] then
				table.remove(Event.Ability.Cooldown.End, i)
				break
			end
		end
		gcdOn = false
	end
end

-- registers a module (castbar, gcd bar) for updates
function bCastUpdateStart(toUpdate)
	local found = false
	for k,v in pairs(updateQueue) do
		-- need to check for duplicates
		if v == toUpdate then
			found = true
			break
		end
	end
	if not found then
		table.insert(updateQueue, toUpdate)
	end
end

-- removes a module from the update queue
function bCastUpdateStop(toStop)
	for i=1,#updateQueue do
		if toStop == updateQueue[i] then
			table.remove(updateQueue, i)
		end
	end
end

-- goes through all modules on the queue and updates them
local function Update()
	for k,v in pairs(updateQueue) do
		-- Loop through all enabled modules, and update them
		v:Update()
	end
end

local function ResetBars(arg)
	-- called on event.unit.castbar
	-- ie. when any unit finishes or starts a cast
	-- loop through all enabled bars to reset them between casts
	for k,v in pairs(enabledBars) do
		-- loop through all enabled cast bars
		if arg[Inspect.Unit.Lookup(v.unit)] then
			-- check if current unit exists in the event table, also if cast starts (value is true)
			-- no need to do anything if cast ends (value is false), or current unit isn't in the table (value is nil)
			-- arg[unitID] = true/false
			v:ResetValues()
		end
	end
end

-- loops through all bars to resize them, fix the fadeout color and to see which ones are enabled
function bCastSettingsChange()
	local bars = {}
	for k,v in pairs(cBars) do
		v:UpdateFrames()
	 	v:FadeOutColor()
		if bCast_profiles[bCast_currentProfile][v.unit].enabled then
			bars[v.unit] = v
			v:Lock(locked)
			bCastUpdateStart(v)
		else
			bCastUpdateStop(v)
		end
		v:SetVisible(false)
	end
	enabledBars = bars
	gcd:UpdatePosition()
	GCDEnabler()
end

function bCastLockToggle(value)
	locked = value
	bCastSettingsChange()
	for k,v in pairs(enabledBars) do
		v:Lock(locked)
	end
end

function bCastIsLocked()
	return locked
end

local function PrintHelp()
	print("--- bCast version " .. bCastVersion .. " help ---")
	print("\'/bcast help\' - Prints this help.")
	print("\'/bcast config\' - Opens the options window.")
	print("\'/bcast profile <profilename>\' - Changes the current profile.")
	print("\'/bcast lock\' - Toggles the locking & unlocking of the cast bars, allowing them to be moved.")
	print("\'/bcast reset\' - Resets the current profile to default values.")
	print("\'/bcast gcd <position>\' - Changes the global cooldown bar position, valid values are \'above\', \'below\', \'none\'.")
	print("")
	print("For bar settings, you need to specify which bar you want to change with its unit name.")
	print("Valid unit names are: \'player\', \'player.target\', \'player.target.target\' and \'focus\'.")
	print("Any command without a new value will print out the current setting.")
	print("\'/bcast <unit> width <new width>\' - Changes cast bar width.")
	print("\'/bcast <unit> height <new width>\' - Changes cast bar height.")
	print("\'/bcast <unit> texture <new texture>\' - Changes cast bar texture, texture file needs to be in bcast\\media\\ folder.")
	print("\'/bcast <unit> fontSize <new size>\' - Changes font size.")
	print("\'/bcast <unit> width <new width>\' - Changes cast bar width.")
	print("\'/bcast <unit> font <new font>\' - Changes the font, font file needs to be in bcast\\media\\ folder.")
	print("\'/bcast <unit> fontColor <R G B>\' - Changes the font color in RGB, range 0 - 255.")
	print("\'/bcast <unit> icon <icon position>\' - Changes the cast icon position, valid values are \'left\', \'right\' and \'none\'.")
	print("\'/bcast <unit> barColor <R G B>\' - Changes the bar color in RGB, range 0 - 255, for example, default bar color would be \'51 201 255\'.")
	print("\'/bcast <unit> unIntBarColor <R G B>\' - Changes the uninterruptible bar color in RGB, range 0 - 255. Doesn\'t work for the player bar.")
	print("\'/bcast <unit> timeStyle <style>\' - Changes if the cast time is shown as time already spent casting (\'done\') or time still left in the cast (\'remaining\').")
	print("\'/bcast <unit> pushback\' - Toggles the pushback display on maximum cast time, only works on the player bar.")
	print("\'/bcast <unit> enable\' - Enables the bar.")
	print("\'/bcast <unit> disable\' - Disables the bar.")
end

local function SlashHandler(args)
	local argz = {}
	local totala = 0
	for cArg in string.gmatch(args, "[^%s]+") do
		argz[totala] = cArg
		totala=totala+1
	end
	
	if totala > 0 then
		if argz[0] == "help" then
			PrintHelp()
		elseif argz[0] == "gcd" then
			if totala >= 1 then
				if argz[1] == "above" or argz[1] == "below" then
					bCast_profiles[bCast_currentProfile].player.showGCD = argz[1]
					GCDEnabler()
					print("GCD display set to " .. argz[1])
					bCastSettingsChange()
				elseif argz[1] == "none" then
					bCast_profiles[bCast_currentProfile].player.showGCD = argz[1]
					GCDEnabler()
					print("GCD display set to " .. argz[1])
					bCastSettingsChange()
				else
					print("Unknown command. Use /bcast help for command list.")
				end
			else
				print("GCD display is set to " .. bCast_profiles[bCast_currentProfile].player.showGCD)
			end
		elseif argz[0] == "config" then
			bCastOptions.Show()
		elseif argz[0] == "profile" then
			if totala > 1 then
				local found = false
				for k,v in pairs(bCast_profiles) do
					if argz[1] == k then
						found = true
						break
					end
				end
				if found then
					if not bCastIsLocked() then bCastLockToggle(true) end
					bCast_currentProfile = argz[1]
					bCastSettingsChange()
					print("Profile changed to " .. argz[1])
				else
					print("Unknown profile")
				end
			else
				print("Current profile is " .. bCast_currentProfile)
			end
		elseif argz[0] == "lock" then
			bCastLockToggle(not bCastIsLocked())
			if bCastIsLocked() then
				print("Bars locked")
			else
				print("Bars unlocked")
			end
		elseif argz[0] == "reset" then
			bCast_profiles[bCast_currentProfile] = bCastDefaults()
			bCastSettingsChange()
			print("Profile reset")
		elseif argz[0] == "player" or argz[0] == "player.target" or argz[0] == "player.target.target" or argz[0] == "focus" then
			-- All the commands for bar settings go here
			if totala > 1 then
				if argz[1] == "width" then
					if totala >= 3 then
						local newW = tonumber(argz[2])
						bCast_profiles[bCast_currentProfile][argz[0]].width = newW
						print(argz[0] .. " width set to " .. newW)
					else
						print("Current width for " .. argz[0] .. " is " .. tostring(bCast_profiles[bCast_currentProfile][argz[0]].width))
					end
				elseif argz[1] == "height" then
					if totala >= 3 then
						local newH = tonumber(argz[2])
						bCast_profiles[bCast_currentProfile][argz[0]].height = newH
						print(argz[0] .. " height set to " .. newH)
					else
						print("Current height for " .. argz[0] .. " is " .. tostring(bCast_profiles[bCast_currentProfile][argz[0]].height))
					end
				elseif argz[1] == "texture" then
					if totala >= 3 then
						bCast_profiles[bCast_currentProfile][argz[0]].texture = argz[2]
						print(argz[0] .. " texture set to " .. argz[2])
					else
						print("Current texture for " .. argz[0] .. " is " .. bCast_profiles[bCast_currentProfile][argz[0]].texture)
					end
				elseif argz[1] == "fontSize" then
					if totala >= 3 then
						bCast_profiles[bCast_currentProfile][argz[0]].fontSize = tonumber(argz[2])
						print(argz[0] .. " font size set to " .. tostring(argz[2])	)
					else
						print("Current font size for " .. argz[0] .. " is " .. tostring(bCast_profiles[bCast_currentProfile][argz[0]].fontSize))
					end
				elseif argz[1] == "font" then
					if totala >= 3 then
						bCast_profiles[bCast_currentProfile][argz[0]].font = argz[2]
						print(argz[0] .. " font set to " .. argz[2])
					else
						print("Current font for " .. argz[0] .. " is " .. bCast_profiles[bCast_currentProfile][argz[0]].font)
					end
				elseif argz[1] == "icon" then
					if totala >=3 then
						if argz[2] == "left" or argz[2] == "right" or argz[2] == "none" then
							bCast_profiles[bCast_currentProfile][argz[0]].icon = argz[2]
							print(argz[0] .. " icon set to " .. argz[2])
						else
							print("Unknown command. Use /bcast help for command list.")
						end
					else
						print("Current icon setting for " .. argz[0] .. " is " .. bCast_profiles[bCast_currentProfile][argz[0]].icon)
					end
				elseif argz[1] == "barColor" then
					if totala >= 5 then
						bCast_profiles[bCast_currentProfile][argz[0]].colors.bar_r = tonumber(argz[2])/255
						bCast_profiles[bCast_currentProfile][argz[0]].colors.bar_g = tonumber(argz[3])/255
						bCast_profiles[bCast_currentProfile][argz[0]].colors.bar_b = tonumber(argz[4])/255
						print(argz[0] .. " bar color set to " .. tostring(math.floor(bCast_profiles[bCast_currentProfile][argz[0]].colors.bar_r*255 + 0.5)) .. " " ..  tostring(math.floor(bCast_profiles[bCast_currentProfile][argz[0]].colors.bar_g*255 + 0.5)) .. " " .. tostring(math.floor(bCast_profiles[bCast_currentProfile][argz[0]].colors.bar_b*255 + 0.5)))
					else
						print("Current bar color for " .. argz[0] .. " is " .. tostring(math.floor(bCast_profiles[bCast_currentProfile][argz[0]].colors.bar_r*255 + 0.5)) .. " " ..  tostring(math.floor(bCast_profiles[bCast_currentProfile][argz[0]].colors.bar_g*255 + 0.5)) .. " " .. tostring(math.floor(bCast_profiles[bCast_currentProfile][argz[0]].colors.bar_b*255 + 0.5)))
					end
				elseif argz[1] == "truncate" then
					if totala >=3 then
						bCast_profiles[bCast_currentProfile][argz[0]].maxLength = tonumber(argz[2])
						print(argz[0] .. " maximum ability name length set to " .. bCast_profiles[bCast_currentProfile][argz[0]].maxLength)
					else
						print("Current maximum ability name length for " .. argz[0] .. " is ".. bCast_profiles[bCast_currentProfile][argz[0]].maxLength)
					end
				elseif argz[1] == "timeStyle" then
					if totala >= 3 then
						if argz[2] == "done" or argz[2] == "remaining" then
							bCast_profiles[bCast_currentProfile][argz[0]].timeStyle = argz[2]
							print(argz[0] .. " time style set to " ..  bCast_profiles[bCast_currentProfile][argz[0]].timeStyle)
						else
							print("Unknown command. Use /bcast help for command list.")
						end
					else
						print("Current time style for " .. argz[0] .. " is " .. bCast_profiles[bCast_currentProfile][argz[0]].timeStyle)
					end
				elseif argz[1] == "enable" then
					 bCast_profiles[bCast_currentProfile][argz[0]].enabled = true
					 print(argz[0] .. " bar enabled")
				elseif argz[1] == "disable" then
					 bCast_profiles[bCast_currentProfile][argz[0]].enabled = false
					 print(argz[0] .. " bar disabled")
				elseif argz[1] == "pushback" then
					if bCast_profiles[bCast_currentProfile][argz[0]].pushBack ~= nil then
						bCast_profiles[bCast_currentProfile][argz[0]].pushBack = not bCast_profiles[bCast_currentProfile][argz[0]].pushBack
						print(argz[0] .. " pushback display set to " .. tostring(bCast_profiles[bCast_currentProfile][argz[0]].pushBack))
					end
				elseif argz[1] == "unIntBarColor" then
					if argz[0] ~= "player" then
						if totala >= 5 then
							bCast_profiles[bCast_currentProfile][argz[0]].colors.bar_ir = tonumber(argz[2])/255
							bCast_profiles[bCast_currentProfile][argz[0]].colors.bar_ig = tonumber(argz[3])/255
							bCast_profiles[bCast_currentProfile][argz[0]].colors.bar_ib = tonumber(argz[4])/255
							print(argz[0] .. " uninterruptible bar color set to " .. tostring(math.floor(bCast_profiles[bCast_currentProfile][argz[0]].colors.bar_ir*255 + 0.5)) .. " " ..  tostring(math.floor(bCast_profiles[bCast_currentProfile][argz[0]].colors.bar_ig*255 + 0.5)) .. " " .. tostring(math.floor(bCast_profiles[bCast_currentProfile][argz[0]].colors.bar_ib*255 + 0.5)))
						else
							print("Current uninterruptible bar color for " .. argz[0] .. " is " .. tostring(math.floor(bCast_profiles[bCast_currentProfile][argz[0]].colors.bar_ir*255 + 0.5)) .. " " ..  tostring(math.floor(bCast_profiles[bCast_currentProfile][argz[0]].colors.bar_ig*255 + 0.5)) .. " " .. tostring(math.floor(bCast_profiles[bCast_currentProfile][argz[0]].colors.bar_ib*255 + 0.5)))
						end
					end
				elseif argz[1] == "fontColor" then
					if totala >= 5 then
						bCast_profiles[bCast_currentProfile][argz[0]].colors.text_r = tonumber(argz[2])/255
						bCast_profiles[bCast_currentProfile][argz[0]].colors.text_g = tonumber(argz[3])/255
						bCast_profiles[bCast_currentProfile][argz[0]].colors.text_b = tonumber(argz[4])/255
						print(argz[0] .. " font color set to " .. tostring(math.floor(bCast_profiles[bCast_currentProfile][argz[0]].colors.text_r*255 + 0.5)) .. " " ..  tostring(math.floor(bCast_profiles[bCast_currentProfile][argz[0]].colors.text_g*255 + 0.5)) .. " " .. tostring(math.floor(bCast_profiles[bCast_currentProfile][argz[0]].colors.text_b*255 + 0.5)))
					else
						print("Current font color for " .. argz[0] .. " is " .. tostring(math.floor(bCast_profiles[bCast_currentProfile][argz[0]].colors.text_r*255 + 0.5)) .. " " ..  tostring(math.floor(bCast_profiles[bCast_currentProfile][argz[0]].colors.text_g*255 + 0.5)) .. " " .. tostring(math.floor(bCast_profiles[bCast_currentProfile][argz[0]].colors.text_b*255 + 0.5)))
					end
				else
					print("Unknown command. Use /bcast help for command list.")
				end
			else
				print("Unknown command. Use /bcast help for command list.")
			end
			bCastSettingsChange()
		else
			print("Unknown command. Use /bcast help for command list.")
		end
	else
		print("Unknown command. Use /bcast help for command list.")
	end
end

local function VariablesLoaded(addon)
	if addon == "bCast" then
		-- create all the bars
		-- Ideally, would only create the enabled ones
		table.insert(cBars, bCastBar.New("player"))
		table.insert(cBars, bCastBar.New("focus"))
		table.insert(cBars, bCastBar.New("player.target"))
		table.insert(cBars, bCastBar.New("player.target.target"))
		gcd = bCastGCDBar.New()
		
		CDStartFunc = { CDStart, "bCast", "CDStart" }
		CDEndFunc = { CDEnd, "bCast", "CDEnd" }
		
		GCDEnabler()
		bCastSettingsChange()
		bCastOptions.AddonLoad()
		
		print("bCast ".. bCastVersion .. " loaded.")
	end
end
	
table.insert(Event.Addon.SavedVariables.Load.End, { VariablesLoaded, "bCast", "VariablesLoaded" } )
table.insert(Command.Slash.Register("bcast"), { SlashHandler, "bCast", "SlashHandler" } )
table.insert(Event.System.Update.Begin, { Update, "bCast", "Update" } )
table.insert(Event.Unit.Castbar, { ResetBars, "bCast", "ResetBars"  } )